"""Contains IndexToMarkdown class and entrypoint for indexToMarkdown_cli."""

import argparse
import json
from tabulate import tabulate
from dateutil.parser import isoparse


class IndexToMarkdown:
    """An IndexToMarkdown object."""

    @staticmethod
    def index_to_markdown(indexobj):
        """Convert the index object [indexobj] (dict) to a markdown string and return the markdown string."""
        lines = [
            "<!-- begin index generated by mitreattack-python.navlayers.collections.col_to_md.py -->",
            f"## {indexobj['name']}",
            f"\n{indexobj['description']}\n",
            f"### {indexobj['name']} Collections\n",
        ]

        for collection in indexobj["collections"]:
            lines.append(f"#### {collection['name']}")
            lines.append(f"\n{collection['description']}\n")
            format_versions = list(
                map(
                    lambda v: [
                        f"[{collection['name']} v{v['version']}]({v['url']})",
                        isoparse(v["modified"]).strftime("%d %B %Y"),
                    ],
                    collection["versions"],
                )
            )

            lines.append(tabulate(format_versions, headers=["version", "released"], tablefmt="github"))
            lines.append("")

        lines.append("<!-- end index generated by  mitreattack-python.navlayers.collections.col_to_md.py -->")

        return "\n".join(lines)


def main(args):
    """Entrypoint for indexToMarkdown_cli."""
    with open(args.index, "r", encoding="utf-16") as f:
        index = json.load(f)
        with open(args.output, "w", encoding="utf-16") as f2:
            f2.write(IndexToMarkdown.index_to_markdown(index))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Print a markdown string to std-out representing a collection index")
    parser.add_argument(
        "-i", "--index", type=str, default="index.json", help="the collection index file to convert to markdown"
    )
    parser.add_argument("-o", "--output", type=str, default="index.md", help="markdown output file")
    argv = parser.parse_args()
    main(argv)



"""Contains UsageLayerGenerator class."""

from stix2 import Filter
from itertools import chain
import copy

from mitreattack.navlayers.exporters.matrix_gen import MatrixGen
from mitreattack.navlayers.core.exceptions import BadInput, typeChecker
from mitreattack.navlayers.core.layer import Layer
from mitreattack.navlayers.generators.gen_helpers import remove_revoked_depreciated, get_attack_id, build_data_strings


class UnableToFindStixObject(Exception):
    """Custom exception used when STIX object is not found."""

    pass


class StixObjectIsNotValid(Exception):
    """Custom exception used when STIX objects are found to not be valid."""

    pass


class UsageLayerGenerator:
    """Generates a Layer that shows techniques mapped to an input group, software or mitigation."""

    def __init__(self, source, domain="enterprise", resource=None):
        """Initialize the Generator.

        :param source: Which source to use for data (local, taxii [server], or [remote] ATT&CK Workbench)
        :param domain: Which matrix to use during generation
        :param resource: string path to local STIX data (local) or url of workbench to reach out to (remote)
        """
        self.matrix_handle = MatrixGen(source, resource)
        self.domain = domain
        try:
            self.source_handle = self.matrix_handle.collections[domain]
        except KeyError:
            print(f"[UsageGenerator] - unable to load collection {domain} (current source = {source}).")
            raise BadInput
        self.full_matrix = self.matrix_handle.get_matrix(self.domain)
        self.sources = self.source_handle.query([Filter("type", "=", "x-mitre-data-source")])
        self.components = self.source_handle.query([Filter("type", "=", "x-mitre-data-component")])
        self.source_mapping = build_data_strings(self.sources, self.components)

    def get_stix_object(self, match):
        """Retrieve the stix object for a given string.

        :param match: The string to match on - can be a name, alias, or ATT&CK ID
        :return: the corresponding stix object
        """
        filts = [
            [Filter("name", "=", match)],
            [Filter(match, "in", "aliases")],
            [Filter(match, "in", "x_mitre_aliases")],
            [Filter("external_references.external_id", "=", match)],
            [Filter("id", "=", match)],  # Support data component type objects from sum generator
        ]
        data = list(chain.from_iterable(self.source_handle.query(f) for f in filts))
        data = remove_revoked_depreciated(data)
        if len(data):
            if len(data) > 1:
                print(
                    f"[Usage Generator] - WARNING! Multiple matches found for {match}: [{data}]. Selecting the first "
                    f"one as default."
                )
            return data[0]
        raise UnableToFindStixObject

    def get_matrix_data(self, match_pattern):
        """Retrieve a list of attack-pattern (technique) objects that map to a group, software or mitigation.

        :param match_pattern: Name, associated group/software (alias), or ATT&CK ID.
                              Techniques mapped to the object matching this pattern are returned.```
        """
        obj = self.get_stix_object(match_pattern)
        if obj["type"] == "course-of-action":
            verb = "mitigates"
        elif obj["type"] == "x-mitre-data-component" or obj["type"] == "x-mitre-data-source":
            verb = "detects"
        else:
            verb = "uses"
        related = self.source_handle.relationships(obj["id"], verb, source_only=True)

        out = self.source_handle.query(
            [Filter("type", "=", "attack-pattern"), Filter("id", "in", [r.target_ref for r in related])]
        )
        return remove_revoked_depreciated(out), obj

    def generate_technique_data(self, raw_matches):
        """Generate technique list of dictionary objects (dictionary form of technique listing for a layer).

        :param raw_matches: matching attack-pattern objects
        :return: list of dictionary objects for every technique: score=0 if not in raw_matches, 1 otherwise,
                    description in comments
        """
        shortlist = []
        for match in raw_matches:
            xid = ""
            xphase = ""
            for ref in match.external_references:
                if ref.source_name == "mitre-attack":
                    xid = ref.external_id
            for phase in match.kill_chain_phases:
                if phase.kill_chain_name == "mitre-attack":
                    xphase = phase.phase_name
            shortlist.append((xid, xphase, match.description))
        full_matrix_listing = copy.deepcopy(self.full_matrix)
        construct = list()
        for tactic in full_matrix_listing:
            for tech in tactic.techniques:
                construct.append(
                    dict(techniqueID=tech.id, score=0, tactic=self.matrix_handle.convert(tactic.tactic.name))
                )
            for tech_key in tactic.subtechniques:
                for subtech in tactic.subtechniques[tech_key]:
                    construct.append(
                        dict(techniqueID=subtech.id, score=0, tactic=self.matrix_handle.convert(tactic.tactic.name))
                    )
        for entry in shortlist:
            for tac in construct:
                if entry[0] == tac["techniqueID"] and (entry[1] == "" or entry[1] == tac["tactic"]):
                    tac["score"] = 1
                    tac["comment"] = entry[2]
        return construct

    def generate_layer(self, match):
        """Generate a layer.

        :param match: the pattern to match
        :return: layer object with annotated techniques
        """
        typeChecker(type(self).__name__, match, str, "match")
        raw_data, matched_obj = self.get_matrix_data(match)
        if matched_obj["type"] not in [
            "course-of-action",
            "tool",
            "malware",
            "intrusion-set",
            "x-mitre-data-source",
            "x-mitre-data-component",
        ]:
            print(
                f"Warning: The input match {match} corresponds with an ATT&CK Technique, which is not supported. "
                f"Please provide a group, software, or mitigation instead."
            )
            raise StixObjectIsNotValid
        a_id = get_attack_id(matched_obj)
        processed_listing = self.generate_technique_data(raw_data)
        raw_layer = dict(name=f"{matched_obj['name']} ({matched_obj['id']})", domain=self.domain + "-attack")
        raw_layer["techniques"] = processed_listing
        output_layer = Layer(raw_layer)
        if matched_obj["type"] != "x-mitre-data-component":
            name = matched_obj["name"]
        else:
            name = self.source_mapping[matched_obj["id"]]
        output_layer.description = (
            f"{self.domain.capitalize() if len(self.domain) > 3 else self.domain.upper()} "
            f"techniques used by {name}, ATT&CK {matched_obj['type']} {a_id}"
        )
        return output_layer



"""Helper functions to convert STIX to pandas dataframes."""

import copy
import datetime
import re
from itertools import chain

import numpy as np
import pandas as pd
from loguru import logger
from stix2 import Filter, MemoryStore
from tqdm import tqdm

from mitreattack.constants import MITRE_ATTACK_ID_SOURCE_NAMES


# Lookup module for Platforms - each matrix has a list of possible platforms, and each platform with multiple
#   subplatforms has a corresponding entry. This allows for a pseudo-recursive lookup of subplatforms, as the presence
#   of a platform at the top level of this lookup indicates the existence of subplatforms.
MATRIX_PLATFORMS_LOOKUP = {
    "enterprise-attack": [
        "PRE",
        "Windows",
        "macOS",
        "Linux",
        "Cloud",
        "Office 365",
        "Azure AD",
        "Google Workspace",
        "SaaS",
        "IaaS",
        "Network",
        "Containers",
    ],
    "mobile-attack": ["Android", "iOS"],
    "Cloud": ["Office 365", "Azure AD", "Google Workspace", "SaaS", "IaaS"],
    "ics-attack": [
        "Field Controller/RTU/PLC/IED",
        "Safety Instrumented System/Protection Relay",
        "Control Server",
        "Input/Output Server",
        "Windows",
        "Human-Machine Interface",
        "Engineering Workstation",
        "Data Historian",
    ],
}

TITLE_EXCLUSIONS = ["and"]


def remove_revoked_deprecated(stix_objects):
    """Remove any revoked or deprecated objects from queries made to the data source."""
    # Note we use .get() because the property may not be present in the JSON data. The default is False
    # if the property is not set.
    return list(
        filter(
            lambda x: x.get("x_mitre_deprecated", False) is False and x.get("revoked", False) is False,
            stix_objects,
        )
    )


def filter_platforms(stix_objects, platforms):
    """Filter out any objects that don't have a matching platform to one in 'platforms'."""
    if not platforms:
        return stix_objects

    return list(
        filter(
            lambda x: any(
                platform.lower() in [y.lower() for y in x.get("x_mitre_platforms", [])] for platform in platforms
            ),
            stix_objects,
        )
    )


def format_date(date):
    """Given a date string, return it formatted as %d %B %Y."""
    if isinstance(date, str):
        date = datetime.datetime.strptime(date, "%Y-%m-%dT%H:%M:%S.%fZ")
    return f"{date.strftime('%d')} {date.strftime('%B')} {date.strftime('%Y')}"


def get_citations(objects):
    """Given a list of STIX objects, return a pandas dataframe for the citations on the objects."""
    citations = []
    for sdo in objects:
        if "external_references" in sdo:
            for ref in sdo["external_references"]:
                if (
                    "external_id" not in ref
                    and "description" in ref
                    and not ref["description"].startswith("(Citation: ")
                ):
                    citation = {
                        "reference": ref["source_name"],
                        "citation": ref["description"],
                    }
                    if "url" in ref:
                        citation["url"] = ref["url"]

                    citations.append(citation)

    return pd.DataFrame(citations).drop_duplicates(subset="reference", ignore_index=True)


def parseBaseStix(sdo):
    """Given an SDO, return a dict of field names:values that are common across all ATT&CK STIX types."""
    row = {}
    url = None
    if sdo.get("external_references"):
        if sdo["external_references"][0]["source_name"] in MITRE_ATTACK_ID_SOURCE_NAMES:
            row["ID"] = sdo["external_references"][0]["external_id"]
            url = sdo["external_references"][0]["url"]
    if "name" in sdo:
        row["name"] = sdo["name"]
    if "description" in sdo:
        row["description"] = sdo["description"]
    if url:
        row["url"] = url
    if "created" in sdo:
        row["created"] = format_date(sdo["created"])
    if "modified" in sdo:
        row["last modified"] = format_date(sdo["modified"])
    if "x_mitre_version" in sdo:
        row["version"] = sdo["x_mitre_version"]
    if "x_mitre_contributors" in sdo:
        row["contributors"] = "; ".join(sorted(sdo["x_mitre_contributors"]))
    return row


def techniquesToDf(src, domain):
    """Parse STIX techniques from the given data and return corresponding pandas dataframes.

    :param src: MemoryStore or other stix2 DataSource object holding the domain data
    :param domain: domain of ATT&CK src corresponds to, e.g "enterprise-attack"
    :returns: a lookup of labels (descriptors/names) to dataframes
    """
    techniques = src.query([Filter("type", "=", "attack-pattern")])
    techniques = remove_revoked_deprecated(techniques)
    technique_rows = []

    all_sub_techniques = src.query(
        [
            Filter("type", "=", "relationship"),
            Filter("relationship_type", "=", "subtechnique-of"),
        ]
    )
    all_sub_techniques = MemoryStore(stix_data=all_sub_techniques)

    for technique in tqdm(techniques, desc="parsing techniques"):
        # get parent technique if sub-technique
        subtechnique = "x_mitre_is_subtechnique" in technique and technique["x_mitre_is_subtechnique"]
        if subtechnique:
            subtechnique_of = all_sub_techniques.query([Filter("source_ref", "=", technique["id"])])[0]
            parent = src.get(subtechnique_of["target_ref"])

        # base STIX properties
        row = parseBaseStix(technique)

        # sub-technique properties
        if "kill_chain_phases" not in technique:
            logger.error(
                f"Skipping {technique['external_references'][0]['external_id']} [{technique['id']}] because it does't have kill chain phases"
            )
            continue
        tactic_shortnames = list(map(lambda kcp: kcp["phase_name"], technique["kill_chain_phases"]))
        tactics = list(
            map(
                lambda t: " ".join([x.title() if x not in TITLE_EXCLUSIONS else x for x in t.split("-")]),
                tactic_shortnames,
            )
        )
        row["tactics"] = ", ".join(sorted(tactics))

        if "x_mitre_detection" in technique:
            row["detection"] = technique["x_mitre_detection"]
        if "x_mitre_platforms" in technique:
            row["platforms"] = ", ".join(sorted(technique["x_mitre_platforms"]))

        # domain specific fields -- ICS + Enterprise
        if domain in ["enterprise-attack", "ics-attack"]:
            if "x_mitre_data_sources" in technique:
                row["data sources"] = ", ".join(sorted(technique["x_mitre_data_sources"]))

        # domain specific fields -- enterprise
        if domain == "enterprise-attack":
            row["is sub-technique"] = subtechnique
            if subtechnique:
                row["name"] = f"{parent['name']}: {technique['name']}"
                row["sub-technique of"] = parent["external_references"][0]["external_id"]

            if "x_mitre_system_requirements" in technique:
                row["system requirements"] = ", ".join(sorted(technique["x_mitre_system_requirements"]))
            if "x_mitre_permissions_required" in technique:
                row["permissions required"] = ", ".join(
                    sorted(technique["x_mitre_permissions_required"], key=str.lower)
                )
            if "x_mitre_effective_permissions" in technique:
                row["effective permissions"] = ", ".join(
                    sorted(technique["x_mitre_effective_permissions"], key=str.lower)
                )

            if "defense-evasion" in tactic_shortnames and "x_mitre_defense_bypassed" in technique:
                row["defenses bypassed"] = ", ".join(sorted(technique["x_mitre_defense_bypassed"]))
            if "execution" in tactic_shortnames and "x_mitre_remote_support" in technique:
                row["supports remote"] = technique["x_mitre_remote_support"]
            if "impact" in tactic_shortnames and "x_mitre_impact_type" in technique:
                row["impact type"] = ", ".join(sorted(technique["x_mitre_impact_type"]))
            capec_refs = list(
                filter(
                    lambda ref: ref["source_name"] == "capec",
                    technique["external_references"],
                )
            )
            if capec_refs:
                row["CAPEC ID"] = ", ".join([x["external_id"] for x in capec_refs])

        # domain specific fields -- mobile
        elif domain == "mobile-attack":
            if "x_mitre_tactic_type" in technique:
                row["tactic type"] = ", ".join(sorted(technique["x_mitre_tactic_type"]))
            mtc_refs = list(
                filter(
                    lambda ref: ref["source_name"] == "NIST Mobile Threat Catalogue",
                    technique["external_references"],
                )
            )
            if mtc_refs:
                row["MTC ID"] = mtc_refs[0]["external_id"]

        technique_rows.append(row)

    citations = get_citations(techniques)
    dataframes = {
        "techniques": pd.DataFrame(technique_rows).sort_values("name"),
    }
    # add relationships
    codex = relationshipsToDf(src, relatedType="technique")
    dataframes.update(codex)
    # add relationship references
    dataframes["techniques"]["relationship citations"] = _get_relationship_citations(dataframes["techniques"], codex)
    # add/merge citations
    if not citations.empty:
        if "citations" in dataframes:  # append to existing citations from references
            dataframes["citations"] = pd.concat([dataframes["citations"], citations])
        else:  # add citations
            dataframes["citations"] = citations

        dataframes["citations"].sort_values("reference")

    return dataframes


def tacticsToDf(src):
    """Parse STIX tactics from the given data and return corresponding pandas dataframes.

    :param src: MemoryStore or other stix2 DataSource object holding the domain data
    :returns: a lookup of labels (descriptors/names) to dataframes
    """
    tactics = src.query([Filter("type", "=", "x-mitre-tactic")])
    tactics = remove_revoked_deprecated(tactics)

    tactic_rows = []
    for tactic in tqdm(tactics, desc="parsing mitigations"):
        tactic_rows.append(parseBaseStix(tactic))

    citations = get_citations(tactics)
    dataframes = {
        "tactics": pd.DataFrame(tactic_rows).sort_values("name"),
    }
    if not citations.empty:
        dataframes["citations"] = citations.sort_values("reference")

    return dataframes


def datasourcesToDf(src):
    """Parse STIX Data Sources and their Data components from the given data and return corresponding pandas dataframes.

    :param src: MemoryStore or other stix2 DataSource object holding the domain data
    :returns: a lookup of labels (descriptors/names) to dataframes
    """
    data = list(
        chain.from_iterable(  # software are the union of the tool and malware types
            src.query(f)
            for f in [
                Filter("type", "=", "x-mitre-data-component"),
                Filter("type", "=", "x-mitre-data-source"),
            ]
        )
    )
    dataframes = {}
    if data:
        refined = remove_revoked_deprecated(data)
        data_object_rows = []
        source_lookup = dict()
        for x in refined:
            if x["type"] == "x-mitre-data-source":
                source_lookup[x["id"]] = x["name"]
        for data_object in tqdm(refined, desc="parsing data sources"):
            # add common STIx fields
            row = parseBaseStix(data_object)
            # add software-specific fields
            if "x_mitre_platforms" in data_object:
                row["platforms"] = ", ".join(sorted(data_object["x_mitre_platforms"]))
            if "x_mitre_collection_layers" in data_object:
                row["collection layers"] = ", ".join(sorted(data_object["x_mitre_collection_layers"]))
            if "x_mitre_aliases" in data_object:
                row["aliases"] = ", ".join(sorted(data_object["x_mitre_aliases"][1:]))
            if data_object["type"] == "x-mitre-data-component":
                row["name"] = f"{source_lookup[data_object['x_mitre_data_source_ref']]}: {data_object['name']}"
                row["type"] = "datacomponent"
            else:
                row["type"] = "datasource"
            if "description" in data_object:
                row["description"] = data_object["description"]
            data_object_rows.append(row)

        citations = get_citations(refined)
        tempa = pd.DataFrame(data_object_rows).sort_values("name")
        dataframes["datasources"] = tempa.reindex(
            columns=[
                "name",
                "ID",
                "description",
                "collection layers",
                "platforms",
                "created",
                "modified",
                "type",
                "version",
                "url",
                "contributors",
            ]
        )
        # add relationships
        dataframes.update(relationshipsToDf(src, relatedType="datasource"))
        # add/merge citations
        if not citations.empty:
            if "citations" in dataframes:  # append to existing citations from references
                dataframes["citations"] = pd.concat([dataframes["citations"], citations])
            else:  # add citations
                dataframes["citations"] = citations

            dataframes["citations"].sort_values("reference")
    else:
        logger.warning("No data components or data sources found - nothing to parse")

    return dataframes


def softwareToDf(src):
    """Parse STIX software from the given data and return corresponding pandas dataframes.

    :param src: MemoryStore or other stix2 DataSource object holding the domain data
    :returns: a lookup of labels (descriptors/names) to dataframes
    """
    software = list(
        chain.from_iterable(  # software are the union of the tool and malware types
            src.query(f) for f in [Filter("type", "=", "tool"), Filter("type", "=", "malware")]
        )
    )
    software = remove_revoked_deprecated(software)
    software_rows = []
    for soft in tqdm(software, desc="parsing software"):
        # add common STIx fields
        row = parseBaseStix(soft)
        # add software-specific fields
        if "x_mitre_platforms" in soft:
            row["platforms"] = ", ".join(sorted(soft["x_mitre_platforms"]))
        if "x_mitre_aliases" in soft:
            row["aliases"] = ", ".join(sorted(soft["x_mitre_aliases"][1:]))
        row["type"] = soft["type"]  # malware or tool

        software_rows.append(row)

    citations = get_citations(software)
    dataframes = {
        "software": pd.DataFrame(software_rows).sort_values("name"),
    }
    # add relationships
    codex = relationshipsToDf(src, relatedType="software")
    dataframes.update(codex)
    # add relationship references
    dataframes["software"]["relationship citations"] = _get_relationship_citations(dataframes["software"], codex)
    # add/merge citations
    if not citations.empty:
        if "citations" in dataframes:  # append to existing citations from references
            dataframes["citations"] = pd.concat([dataframes["citations"], citations])
        else:  # add citations
            dataframes["citations"] = citations

        dataframes["citations"].sort_values("reference")

    return dataframes


def groupsToDf(src):
    """Parse STIX groups from the given data and return corresponding pandas dataframes.

    :param src: MemoryStore or other stix2 DataSource object holding the domain data
    :returns: a lookup of labels (descriptors/names) to dataframes
    """
    groups = src.query([Filter("type", "=", "intrusion-set")])
    groups = remove_revoked_deprecated(groups)
    group_rows = []
    for group in tqdm(groups, desc="parsing groups"):
        row = parseBaseStix(group)

        # add group aliases
        if "aliases" in group:
            associated_groups = []
            associated_group_citations = []
            for alias in sorted(group["aliases"][1:]):
                # find the reference for the alias
                associated_groups.append(alias)
                for ref in group["external_references"]:
                    if ref["source_name"] == alias:
                        associated_group_citations.append(ref["description"])
                        break
                        # aliases.append(alias)
            row["associated groups"] = ", ".join(associated_groups)
            row["associated groups citations"] = ", ".join(associated_group_citations)

        group_rows.append(row)

    citations = get_citations(groups)
    dataframes = {
        "groups": pd.DataFrame(group_rows).sort_values("name"),
    }
    # add relationships
    codex = relationshipsToDf(src, relatedType="group")
    dataframes.update(codex)
    # add relationship references
    dataframes["groups"]["relationship citations"] = _get_relationship_citations(dataframes["groups"], codex)
    # add/merge citations
    if not citations.empty:
        # append to existing citations from references
        if "citations" in dataframes:
            dataframes["citations"] = pd.concat([dataframes["citations"], citations])
        else:  # add citations
            dataframes["citations"] = citations

        dataframes["citations"].sort_values("reference")

    return dataframes


def campaignsToDf(src):
    """Parse STIX campaigns from the given data and return corresponding pandas dataframes.

    :param src: MemoryStore or other stix2 DataSource object holding the domain data
    :returns: a lookup of labels (descriptors/names) to dataframes
    """
    campaigns = src.query([Filter("type", "=", "campaign")])
    campaigns = remove_revoked_deprecated(campaigns)

    dataframes = {}
    if campaigns:
        campaign_rows = []
        for campaign in tqdm(campaigns, desc="parsing campaigns"):
            row = parseBaseStix(campaign)

            # add group aliases
            if "aliases" in campaign:
                associated_campaigns = []
                associated_campaign_citations = []
                for alias in sorted(campaign["aliases"][1:]):
                    # find the reference for the alias
                    associated_campaigns.append(alias)
                    for ref in campaign["external_references"]:
                        if ref["source_name"] == alias:
                            associated_campaign_citations.append(ref["description"])
                            break
                            # aliases.append(alias)
                row["associated campaigns"] = ", ".join(associated_campaigns)
                row["associated campaigns citations"] = ", ".join(associated_campaign_citations)

            campaign_rows.append(row)

        citations = get_citations(campaigns)
        dataframes = {
            "campaigns": pd.DataFrame(campaign_rows).sort_values("name"),
        }
        # add relationships
        codex = relationshipsToDf(src, relatedType="campaign")
        dataframes.update(codex)

        # add relationship references
        dataframes["campaigns"]["relationship citations"] = _get_relationship_citations(dataframes["campaigns"], codex)

        # add/merge citations
        if not citations.empty:
            # append to existing citations from references
            if "citations" in dataframes:
                dataframes["citations"] = pd.concat([dataframes["citations"], citations])
            else:
                dataframes["citations"] = citations

            dataframes["citations"].sort_values("reference")
    else:
        logger.warning("No campaigns found - nothing to parse")

    return dataframes


def mitigationsToDf(src):
    """Parse STIX mitigations from the given data and return corresponding pandas dataframes.

    :param src: MemoryStore or other stix2 DataSource object holding the domain data
    :returns: a lookup of labels (descriptors/names) to dataframes
    """
    mitigations = src.query([Filter("type", "=", "course-of-action")])
    mitigations = remove_revoked_deprecated(mitigations)
    mitigation_rows = []
    for mitigation in tqdm(mitigations, desc="parsing mitigations"):
        mitigation_rows.append(parseBaseStix(mitigation))

    citations = get_citations(mitigations)
    dataframes = {
        "mitigations": pd.DataFrame(mitigation_rows).sort_values("name"),
    }
    # add relationships
    codex = relationshipsToDf(src, relatedType="mitigation")
    dataframes.update(codex)
    # add relationship references
    dataframes["mitigations"]["relationship citations"] = _get_relationship_citations(dataframes["mitigations"], codex)
    # add/merge citations
    if not citations.empty:
        if "citations" in dataframes:  # append to existing citations from references
            dataframes["citations"] = pd.concat([dataframes["citations"], citations])
        else:  # add citations
            dataframes["citations"] = citations

        dataframes["citations"].sort_values("reference")

    return dataframes


class CellRange:
    """Helper class for handling ranges of cells in a spreadsheet. Note: not 0-indexed, row and cols start at 1.

    Data is optional argument for data to store in the cellrange in the case of merged ranges
    format is a dict {name, format} for the XlsxWriter style. Formats of the same name will not be defined multiple
    times to the worksheet; only the first definition will be used
    """

    def __init__(self, leftCol, rightCol, topRow, bottomRow, data=None, format=None):
        self.leftCol = leftCol
        self.rightCol = rightCol
        self.topRow = topRow
        self.bottomRow = bottomRow
        self.data = data
        self.format = format

    def to_excel_format(self):
        """Return the range in excel format, e.g A4:C7."""
        return f"{self._loc_to_excel(self.topRow, self.leftCol)}:{self._loc_to_excel(self.bottomRow, self.rightCol)}"

    def _loc_to_excel(self, row, col):
        """Convert given row and column number to an Excel-style cell name. Note: not 0-indexed."""
        letters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        result = []
        while col:
            col, rem = divmod(col - 1, 26)
            result[:0] = letters[rem]
        return "".join(result) + str(row)


def build_technique_and_sub_columns(
    src, techniques, columns, merge_data_handle, matrix_grid_handle, tactic_name, platform=None
):
    """Build technique and subtechnique columns for a given matrix and attach them to the appropriate object listings.

    :param src: MemoryStore or other stix2 DataSource object holding the domain data
    :param techniques: List of technique stix objects belong in this tactic column
    :param columns: Existing columns in this matrix (used for placement)
    :param merge_data_handle: Handle to the 'merge' data object for this matrix
    :param matrix_grid_handle: Handle to the 2D grid array being constructed for the matrix (technique and subtechnique
                                columns will be appended here)
    :param tactic_name: The name of the corresponding tactic for this column
    :param platform: [Optional] The name of a platform to filter subtechniques by

    :return: Nothing (meta - modifies the passed in merge_data_handle and matrix_grid_handle objects)
    """
    techniques_column = []
    subtechniques_column = []

    all_sub_techniques = src.query(
        [
            Filter("type", "=", "relationship"),
            Filter("relationship_type", "=", "subtechnique-of"),
        ]
    )
    all_sub_techniques = MemoryStore(stix_data=all_sub_techniques)

    for technique in techniques:
        techniques_column.append(technique["name"])

        # sub-technique relationships
        subtechnique_ofs = all_sub_techniques.query([Filter("target_ref", "=", technique["id"])])

        # if there are sub-techniques on the tactic
        if len(subtechnique_ofs) > 0:
            # top of row range to merge
            technique_top = len(techniques_column) + 1

            subtechniques = [src.get(rel["source_ref"]) for rel in subtechnique_ofs]
            if platform:
                subtechniques = filter_platforms(
                    subtechniques,
                    MATRIX_PLATFORMS_LOOKUP[platform] if platform in MATRIX_PLATFORMS_LOOKUP else [platform],
                )

            subtechniques = remove_revoked_deprecated(subtechniques)
            subtechniques = sorted(subtechniques, key=lambda x: x["name"])
            for i in range(len(subtechniques)):  # for each sub-technique
                if i != 0:
                    techniques_column.append("")  # first sub-technique is parallel to the technique in the layout
                subtechniques_column.append(subtechniques[i]["name"])
            technique_bottom = len(techniques_column) + 1  # bottom of row range to merge
            if technique_top != technique_bottom:  # more than 1 sub-technique
                merge_data_handle.append(
                    CellRange(  # merge technique portion of cell group
                        len(columns),
                        len(columns),
                        technique_top,
                        technique_bottom,
                        data=technique["name"],
                        format={  # format of the merged range
                            "name": "supertechnique",
                            "format": {
                                "valign": "vcenter",
                                "text_wrap": 1,
                                "shrink": 1,
                            },
                        },
                    )
                )

        # no sub-techniques; add empty cell parallel to technique
        else:
            subtechniques_column.append("")
    # end adding techniques and sub-techniques to column

    # add technique column to grid
    matrix_grid_handle.append(techniques_column)

    # if there are sub-techniques for the tactic
    if len(list(filter(lambda x: x != "", subtechniques_column))) > 0:
        # add sub-technique sub-column
        matrix_grid_handle.append(subtechniques_column)

        # add empty tactic header for the sub-column
        columns.append("")

        # merge tactic column header with the sub-column header that was just appended
        merge_data_handle.append(
            CellRange(
                len(columns) - 1,
                len(columns),
                1,
                1,
                data=tactic_name,
                format={  # tactic header formatting
                    "name": "tacticHeader",
                    "format": {
                        "bold": 1,
                        "border": 1,
                        "font_size": 14,
                        "align": "center",
                        "shrink": 1,
                    },
                },
            )
        )


def build_parsed_DF_matrix(matrix_grid, columns, merge, parsed_dict):
    """Build the DF matrix object.

    :param matrix_grid: 2D array of the matrix to build
    :param columns: Column headers
    :param merge: Any applicable cell merge ranges and styles
    :param parsed_dict: Dictionary containing name and description for the matrix
    :return: { matrix, name, description, merge, border } where
        matrix is a pandas dataframe of the matrix
        name is the name of the matrix
        description is the description of the matrix
        merge is a list of CellRange objects that need to be merged for formatting of the sub-techniques in the matrix
        columns is the number of columns in the data
    """
    parsed = copy.deepcopy(parsed_dict)
    # reshape array so that pandas consumes it properly
    matrix_grid = np.flip(np.rot90(matrix_grid), 0)
    # create dataframe for array
    df = pd.DataFrame(matrix_grid, columns=columns)

    # Set additional data for the matrix
    parsed["matrix"] = df  # actual dataframe
    parsed["merge"] = merge  # merge ranges and associated styles
    parsed["columns"] = len(columns)  # number of columns with data
    return parsed


def matricesToDf(src, domain):
    """Parse STIX matrices from the given data and return parsed matrix structures.

    :param src: MemoryStore or other stix2 DataSource object holding the domain data
    :param domain: domain of ATT&CK src corresponds to, e.g "enterprise-attack"
    :returns: [{ matrix, name, description, merge, border }, ... ] where
        matrix is a pandas dataframe of the matrix
        name is the name of the matrix
        description is the description of the matrix
        merge is a list of CellRange objects that need to be merged for formatting of the sub-techniques in the matrix
        columns is the number of columns in the data
    """
    matrices = src.query([Filter("type", "=", "x-mitre-matrix")])
    matrices = remove_revoked_deprecated(matrices)
    matrices_parsed = []
    sub_matrices_parsed = []

    for matrix in tqdm(matrices, desc="parsing matrices"):
        sub_matrices_grid = dict()
        sub_matrices_merges = dict()
        sub_matrices_columns = dict()
        for entry in MATRIX_PLATFORMS_LOOKUP[domain]:
            sub_matrices_grid[entry] = []
            sub_matrices_merges[entry] = []
            sub_matrices_columns[entry] = []

        parsed = {
            "name": matrix["name"] if len(matrices) == 1 else f"{domain.split('-')[0].capitalize()} {matrix['name']}",
            "description": matrix["description"],
        }

        matrix_grid = []  # matrix layout in 2d array
        merge = []  # list of CellRange objects to merge later

        columns = []  # column names
        for tactic_ref in tqdm(matrix["tactic_refs"], desc="processing matrix tactics"):
            tactic = src.get(tactic_ref)
            columns.append(tactic["name"])  # add tactic header

            # parse techniques in tactic
            techniques = list(
                filter(
                    lambda t: not ("x_mitre_is_subtechnique" in t and t["x_mitre_is_subtechnique"]),
                    src.query(
                        [
                            Filter("type", "=", "attack-pattern"),
                            Filter(
                                "kill_chain_phases.phase_name",
                                "=",
                                tactic["x_mitre_shortname"],
                            ),
                        ]
                    ),
                )
            )
            techniques = remove_revoked_deprecated(techniques)
            techniques = sorted(techniques, key=lambda x: x["name"])
            # add techniques
            build_technique_and_sub_columns(
                src=src,
                techniques=techniques,
                columns=columns,
                merge_data_handle=merge,
                matrix_grid_handle=matrix_grid,
                tactic_name=tactic["name"],
            )

            for platform in MATRIX_PLATFORMS_LOOKUP[domain]:
                # In order to support "groups" of platforms, each platform is checked against the lookup a second time.
                # If an second entry can be found, the results from that query will be used, otherwise, the singular
                # platform will be.
                a_techs = filter_platforms(
                    techniques,
                    MATRIX_PLATFORMS_LOOKUP[platform] if platform in MATRIX_PLATFORMS_LOOKUP else [platform],
                )
                if a_techs:
                    sub_matrices_columns[platform].append(tactic["name"])
                    build_technique_and_sub_columns(
                        src=src,
                        techniques=a_techs,
                        columns=sub_matrices_columns[platform],
                        merge_data_handle=sub_matrices_merges[platform],
                        matrix_grid_handle=sub_matrices_grid[platform],
                        tactic_name=tactic["name"],
                        platform=platform,
                    )

        # square the grid because pandas doesn't like jagged columns
        longest_column = 0
        for column in matrix_grid:
            longest_column = max(len(column), longest_column)
        for column in matrix_grid:
            for i in range((longest_column - len(column))):
                column.append("")

        for submatrix in sub_matrices_grid:
            mg = sub_matrices_grid[submatrix]
            for column in mg:
                longest_column = max(len(column), longest_column)
            for column in mg:
                for i in range((longest_column - len(column))):
                    column.append("")
        # matrix is now squared

        parsed = build_parsed_DF_matrix(matrix_grid, columns, merge, parsed)
        matrices_parsed.append(parsed)

        for submatrix in sub_matrices_grid:
            if sub_matrices_grid[submatrix]:  # make sure we found matches for something
                local = copy.deepcopy(parsed)
                local["name"] = f"{submatrix}" if len(matrices) == 1 else f"{submatrix} {matrix['name']}"
                local["description"] = local["description"].split(":")[0] + f": {submatrix}"
                subparsed = build_parsed_DF_matrix(
                    sub_matrices_grid[submatrix],
                    sub_matrices_columns[submatrix],
                    sub_matrices_merges[submatrix],
                    local,
                )
                sub_matrices_parsed.append(subparsed)

    # end adding of matrices
    return matrices_parsed, sub_matrices_parsed


def relationshipsToDf(src, relatedType=None):
    """Parse STIX relationships from the given data and return corresponding pandas dataframes.

    :param src: MemoryStore or other stix2 DataSource object holding the domain data
    :param relatedType: optional, singular attack type to only return relationships with, e.g "mitigation"
    :returns: a lookup of labels (descriptors/names) to dataframes
    """
    # Helper lookups
    attackToStixTerm = {
        "technique": ["attack-pattern"],
        "tactic": ["x-mitre-tactic"],
        "software": ["tool", "malware"],
        "group": ["intrusion-set"],
        "campaign": ["campaign"],
        "mitigation": ["course-of-action"],
        "matrix": ["x-mitre-matrix"],
        "datasource": ["x-mitre-data-component"],
    }
    stixToAttackTerm = {
        "attack-pattern": "technique",
        "x-mitre-tactic": "tactic",
        "tool": "software",
        "malware": "software",
        "intrusion-set": "group",
        "course-of-action": "mitigation",
        "x-mitre-matrix": "matrix",
        "x-mitre-data-component": "datacomponent",
        "x-mitre-data-source": "datasource",
        "campaign": "campaign",
    }

    # get master list of relationships
    relationships = src.query([Filter("type", "=", "relationship")])
    relationships = remove_revoked_deprecated(relationships)
    relationship_rows = []  # build list of rows for dataframe
    # tqdm description depends on the related type and parameters
    iterdesc = "parsing all relationships" if not relatedType else f"parsing relationships for type={relatedType}"
    for relationship in tqdm(relationships, desc=iterdesc):
        source = src.get(relationship["source_ref"])  # source object of the relationship
        target = src.get(relationship["target_ref"])  # target object of the relationship

        # filter if related objects don't exist or are revoked or deprecated
        if not source or source.get("x_mitre_deprecated", False) is True or source.get("revoked", False) is True:
            continue
        if not target or target.get("x_mitre_deprecated", False) is True or target.get("revoked", False) is True:
            continue
        if relationship["relationship_type"] == "revoked":
            continue

        # don't track sub-technique relationships, those are tracked in the techniques df
        if relationship["relationship_type"] == "subtechnique-of":
            continue

        # filter out relationships not with relatedType
        if relatedType:
            related = False

            # try all stix types for the ATT&CK type
            for stixTerm in attackToStixTerm[relatedType]:
                # if any stix type is part of the relationship
                if source["type"] == stixTerm or target["type"] == stixTerm:
                    related = True
                    break

            if not related:
                # skip this relationship if the types don't match
                continue

        # add mapping data
        row = {}

        def add_side(label, sdo):
            """Add data for one side of the mapping."""
            # logger.debug(sdo)
            if sdo.get("external_references"):
                if sdo["external_references"][0]["source_name"] in MITRE_ATTACK_ID_SOURCE_NAMES:
                    # "source ID" or "target ID"
                    row[f"{label} ID"] = sdo["external_references"][0]["external_id"]

            if "name" in sdo:
                # "source name" or "target name"
                row[f"{label} name"] = sdo["name"]

            # "source type" or "target type"
            row[f"{label} type"] = stixToAttackTerm[sdo["type"]]

        add_side("source", source)
        row["mapping type"] = relationship["relationship_type"]  # mapping type goes between the source/target data
        add_side("target", target)
        if "description" in relationship:  # add description of relationship to the end of the row
            row["mapping description"] = relationship["description"]

        relationship_rows.append(row)

    citations = get_citations(relationships)
    relationships = pd.DataFrame(relationship_rows).sort_values(
        ["mapping type", "source type", "target type", "source name", "target name"]
    )

    # return all relationships and citations
    if not relatedType:
        dataframes = {
            "relationships": relationships,
        }
        if not citations.empty:
            dataframes["citations"] = citations.sort_values("reference")

        return dataframes

    # break into dataframes by mapping type
    else:
        dataframes = {}

        relatedGroupSoftware = relationships.query(
            "(`source type` == 'group' or `source type` == 'software') and "
            "`mapping type` == 'uses' and "
            "(`target type` == 'group' or `target type` == 'software')"
        )
        relatedCampaignSoftware = relationships.query(
            "(`source type` == 'campaign' or `source type` == 'software') and "
            "`mapping type` == 'uses' and "
            "(`target type` == 'campaign' or `target type` == 'software')"
        )
        procedureExamples = relationships.query("`mapping type` == 'uses' and `target type` == 'technique'")
        attributedCampaignGroup = relationships.query("`mapping type` == 'attributed-to' and `target type` == 'group'")
        relatedMitigations = relationships.query("`mapping type` == 'mitigates'")

        if not relatedGroupSoftware.empty:
            if relatedType == "group":
                sheet_name = "associated software"
            else:
                sheet_name = "associated groups"
            dataframes[sheet_name] = relatedGroupSoftware

        if not relatedCampaignSoftware.empty:
            if relatedType == "campaign":
                sheet_name = "associated software"
            else:
                sheet_name = "associated campaigns"
            dataframes[sheet_name] = relatedCampaignSoftware

        if not procedureExamples.empty:
            if relatedType == "technique":
                sheet_name = "procedure examples"
            else:
                sheet_name = "techniques used"
            dataframes[sheet_name] = procedureExamples

        if not attributedCampaignGroup.empty:
            if relatedType == "campaign":
                sheet_name = "attributed groups"
            elif relatedType == "group":
                sheet_name = "attributed campaigns"
            else:
                sheet_name = "associated campaigns"
            dataframes[sheet_name] = attributedCampaignGroup

        if not relatedMitigations.empty:
            if relatedType == "technique":
                sheet_name = "associated mitigations"
            else:
                sheet_name = "techniques addressed"
            dataframes[sheet_name] = relatedMitigations

        if not citations.empty:
            # filter citations by ones actually used
            # build master list of used citations
            usedCitations = set()
            for dfname in dataframes:
                df = dataframes[dfname]
                # filter out missing descriptions which for whatever reason
                for description in filter(lambda x: x == x, df["mapping description"].tolist()):
                    # in pandas don't equal themselves
                    [usedCitations.add(x) for x in re.findall(r"\(Citation: (.*?)\)", description)]

            # filter to only used references
            citations = citations[citations.reference.isin(list(usedCitations))]

            dataframes["citations"] = citations.sort_values("reference")

        return dataframes


def _get_relationship_citations(object_dataframe, relationship_df):
    """Extract citations for each _object_ in the relationship dataframe.

    This allows us to include citations from relationships for each ATT&CK object type.

    :param object_dataframe: Dataframe for relevant ATT&CK object
    :param relationship_df: Dataframe of relationships
    :return: Array of strings, with each string being placed relative to the object listing, and containing all
        relevant citations
    """
    object_listing = [x for x in object_dataframe["ID"]]
    new_citations = []
    for z in [x for x in relationship_df if x != "citations"]:
        subset = []
        for y in object_listing:
            mask = relationship_df[z].values == y
            filtered = relationship_df[z].loc[mask]
            temp = set()
            for description in filter(lambda x: x == x, filtered["mapping description"].tolist()):
                [temp.add(x) for x in re.findall(r"\(Citation: (.*?)\)", description)]
            subset.append(",".join([f"(Citation: {z})" for z in temp]))
        if not new_citations:
            new_citations = subset
        else:
            for i in range(0, len(new_citations)):
                new_citations[i] = ",".join([new_citations[i], subset[i]])
    return new_citations



"""Constants used throughout the mitreattack library."""

# The ATT&CK ID entry always has a source_name with a value from:
# ['mitre-attack', 'mitre-mobile-attack', 'mobile-attack', 'mitre-ics-attack']
MITRE_ATTACK_ID_SOURCE_NAMES = ["mitre-attack", "mobile-attack", "mitre-mobile-attack", "mitre-ics-attack"]
MITRE_ATTACK_DOMAIN_STRINGS = ["mitre-attack", "mitre-mobile-attack", "mitre-ics-attack"]



"""Contains Link and LinkDiv classes."""

from mitreattack.navlayers.core.exceptions import typeChecker


class Link:
    """A Link object."""

    def __init__(self, label, url):
        """Initialize - Creates a link object.

        :param label: the label for this link entry
        :param url: the corresponding url for this link entry
        """
        self.label = label
        self.url = url

    @property
    def label(self):
        """Getter for label."""
        return self.__label

    @label.setter
    def label(self, label):
        """Setter for label."""
        typeChecker(type(self).__name__, label, str, "label")
        self.__label = label

    @property
    def url(self):
        """Getter for url."""
        return self.__url

    @url.setter
    def url(self, url):
        """Setter for url."""
        typeChecker(type(self).__name__, url, str, "url")
        self.__url = url

    def get_dict(self):
        """Convert the currently loaded data into a dict.

        :returns: A dict representation of the local metadata object
        """
        return dict(label=self.__label, url=self.__url)


class LinkDiv:
    """A LinkDiv object."""

    def __init__(self, active):
        """Initialize - Creates a Link object divider."""
        self.__name = "DIVIDER"
        self.__value = active

    @property
    def name(self):
        """Getter for name."""
        return self.__name

    @property
    def state(self):
        """Getter for state."""
        return self.__value

    @state.setter
    def state(self, state):
        """Setter for state."""
        typeChecker(type(self).__name__, state, bool, "state")
        self.__value = state

    def get_dict(self):
        """Convert the currently loaded data into a dict.

        :returns: A dict representation of the local metadata object
        """
        return dict(name=self.__name, value=self.__value)



"""Functions to convert ATT&CK STIX data to Excel, as well as entrypoint for attackToExcel_cli."""

import argparse
import os
from typing import Dict, List

import pandas as pd
import requests
from loguru import logger
from stix2 import MemoryStore

# import mitreattack.attackToExcel.stixToDf as stixToDf
from mitreattack.attackToExcel import stixToDf

INVALID_CHARACTERS = ["\\", "/", "*", "[", "]", ":", "?"]
SUB_CHARACTERS = ["\\", "/"]


def get_stix_data(domain: str, version: str = None, remote: str = None, stix_file: str = None) -> MemoryStore:
    """Download the ATT&CK STIX data for the given domain and version from MITRE/CTI (or just domain if a remote workbench is specified).

    Parameters
    ----------
    domain : str
        The domain of ATT&CK to fetch data from, e.g "enterprise-attack"
    version : str, optional
        The version of attack to fetch data from, e.g "v8.1".
        If omitted, returns the latest version (not used for invocations that use remote), by default None
    remote : str, optional
        Optional url to a ATT&CK workbench instance.
        If specified, data will be retrieved from the target Workbench instead of MITRE/CTI, by default None
    stix_file : str, optional
        Path to a local STIX file containing ATT&CK data for a domain, by default None

    Returns
    -------
    MemoryStore
        A stix2.MemoryStore object containing the domain data

    Raises
    ------
    ValueError
        Raised if both `remote` and `stix_file` are passed
    FileNotFoundError
        Raised if `stix_file` not found
    """
    if remote and stix_file:
        raise ValueError("remote and stix_file are mutually exclusive. Please only use one or the other")

    mem_store = None
    if stix_file:
        if os.path.exists(stix_file):
            logger.info(f"Loading STIX file from: {stix_file}")
            mem_store = MemoryStore()
            mem_store.load_from_file(stix_file)
        else:
            raise FileNotFoundError(f"{stix_file} file does not exist.")
    else:
        if remote:
            logger.info("Downloading ATT&CK data from an ATT&CK Workbench instance")
            if ":" not in remote[6:]:
                remote += ":3000"
            if not remote.startswith("http"):
                remote = "http://" + remote
            url = f"{remote}/api/stix-bundles?domain={domain}&includeRevoked=true&includeDeprecated=true"
            stix_json = requests.get(url).json()
            mem_store = MemoryStore(stix_json)
        else:
            logger.info("Downloading ATT&CK data from github.com/mitre/cti")
            if version:
                url = f"https://raw.githubusercontent.com/mitre/cti/ATT%26CK-{version}/{domain}/{domain}.json"
            else:
                url = f"https://raw.githubusercontent.com/mitre/cti/master/{domain}/{domain}.json"

            stix_json = requests.get(url).json()
            mem_store = MemoryStore(stix_data=stix_json["objects"])

    return mem_store


def build_dataframes(src: MemoryStore, domain: str) -> Dict:
    """Build pandas dataframes for each attack type, and return a dictionary lookup for each type to the relevant dataframe.

    :returns:

    Parameters
    ----------
    src : MemoryStore
        MemoryStore or other stix2 DataSource object
    domain : str
        domain of ATT&CK src corresponds to, e.g "enterprise-attack"

    Returns
    -------
    dict
        A dict lookup of each ATT&CK type to dataframes for the given type to be ingested by write_excel
    """
    df = {
        "techniques": stixToDf.techniquesToDf(src, domain),
        "tactics": stixToDf.tacticsToDf(src),
        "software": stixToDf.softwareToDf(src),
        "groups": stixToDf.groupsToDf(src),
        "campaigns": stixToDf.campaignsToDf(src),
        "mitigations": stixToDf.mitigationsToDf(src),
        "matrices": stixToDf.matricesToDf(src, domain),
        "relationships": stixToDf.relationshipsToDf(src),
    }
    # get each ATT&CK type
    if domain == "enterprise-attack":
        df["datasources"] = stixToDf.datasourcesToDf(src)
    return df


def write_excel(dataframes: Dict, domain: str, version: str = None, output_dir: str = ".") -> List:
    """Given a set of dataframes from build_dataframes, write the ATT&CK dataset to output directory.

    Parameters
    ----------
    dataframes : dict
        A dictionary of pandas dataframes as built by build_dataframes()
    domain : str
        Domain of ATT&CK the dataframes correspond to, e.g "enterprise-attack"
    version : str, optional
        The version of ATT&CK the dataframes correspond to, e.g "v8.1".
        If omitted, the output files will not be labelled with the version number, by default None
    output_dir : str, optional
        The directory to write the excel files to.
        If omitted writes to a subfolder of the current directory depending on specified domain and version, by default "."

    Returns
    -------
    list
        A list of filepaths corresponding to the files written by the function
    """
    logger.info("writing formatted files... ")
    # master list of files that have been written
    written_files = []
    # set up output directory
    if version:
        domain_version_string = f"{domain}-{version}"
    else:
        domain_version_string = domain
    output_directory = os.path.join(output_dir, domain_version_string)
    if not os.path.exists(output_directory):
        os.makedirs(output_directory)
    # master dataset file
    master_fp = os.path.join(output_directory, f"{domain_version_string}.xlsx")
    with pd.ExcelWriter(master_fp, engine="xlsxwriter") as master_writer:
        # master list of citations
        citations = pd.DataFrame()

        # write individual dataframes and add to master writer
        for object_type, object_data in dataframes.items():
            fp = os.path.join(output_directory, f"{domain_version_string}-{object_type}.xlsx")

            if object_type != "matrices":
                if not object_data:
                    logger.warning(f"No data for {object_type}. Skipping building an Excel file.")
                    continue

                # write the dataframes for the object type into named sheets
                with pd.ExcelWriter(fp) as object_writer:
                    for sheet_name in object_data:
                        logger.debug(f"Writing sheet to {fp}: {sheet_name}")
                        object_data[sheet_name].to_excel(object_writer, sheet_name=sheet_name, index=False)
                written_files.append(fp)

                # add citations to master citations list
                if "citations" in object_data:
                    citations = pd.concat([citations, object_data["citations"]])

                # add main df to master dataset
                logger.debug(f"Writing sheet to {master_fp}: {object_type}")
                object_data[object_type].to_excel(master_writer, sheet_name=object_type, index=False)

            else:  # handle matrix special formatting
                with pd.ExcelWriter(fp, engine="xlsxwriter") as matrix_writer:
                    # Combine both matrix types
                    combined = object_data[0] + object_data[1]

                    # some domains have multiple matrices
                    for matrix in combined:
                        # name them accordingly if there are multiple
                        sheetname = "matrix" if len(combined) == 1 else matrix["name"] + " matrix"
                        for character in INVALID_CHARACTERS:
                            sheetname = sheetname.replace(character, " or " if character in SUB_CHARACTERS else " ")

                        if len(sheetname) > 31:
                            sheetname = sheetname[0:28] + "..."
                        listing = []

                        # avoid printing subtype matrices to the master file
                        if matrix in object_data[0]:
                            # write unformatted matrix data to master file
                            logger.debug(f"Writing sheet to {master_fp}: {sheetname}")
                            matrix["matrix"].to_excel(master_writer, sheet_name=sheetname, index=False)
                            listing.append(master_writer)

                        # write unformatted matrix to matrix file
                        logger.debug(f"Writing sheet to {fp}: {sheetname}")
                        matrix["matrix"].to_excel(matrix_writer, sheet_name=sheetname, index=False)
                        listing.append(matrix_writer)

                        # for each writer, format the matrix for readability
                        for writer in listing:
                            # define column border styles
                            borderleft = writer.book.add_format({"left": 1, "shrink": 1})
                            borderright = writer.book.add_format({"right": 1, "shrink": 1})

                            # formats only need to be defined once: pointers stored here for subsequent uses
                            formats = {}
                            sheet = writer.sheets[sheetname]

                            # set all columns to 20 width, and add text shrinking to fit
                            sheet.set_column(0, matrix["columns"], width=20)

                            # merge supertechniques and tactic headers if sub-techniques are present on a tactic
                            for merge_range in matrix["merge"]:
                                # sometimes merge ranges have formats to add to the merged range
                                if merge_range.format:
                                    # add format to book if not defined
                                    if merge_range.format["name"] not in formats:
                                        formats[merge_range.format["name"]] = writer.book.add_format(
                                            merge_range.format["format"]
                                        )
                                    # get saved format if already added
                                    theformat = formats[merge_range.format["name"]]

                                    # tactic header merge has additional behavior
                                    if merge_range.format["name"] == "tacticHeader":
                                        # also set border for entire column for grouping
                                        sheet.set_column(
                                            merge_range.leftCol - 1,
                                            merge_range.leftCol - 1,
                                            width=20,  # set column widths to make matrix more readable
                                            cell_format=borderleft,  # left border around tactic
                                        )
                                        sheet.set_column(
                                            merge_range.rightCol - 1,
                                            merge_range.rightCol - 1,
                                            width=20,  # set column widths to make matrix more readable
                                            cell_format=borderright,  # right border around tactic
                                        )
                                else:
                                    theformat = None  # no format

                                # apply the merge
                                sheet.merge_range(merge_range.to_excel_format(), merge_range.data, theformat)

                written_files.append(fp)

        # remove duplicate citations and add sheet to master file
        logger.debug(f"Writing sheet to {master_fp}: citations")
        citations.drop_duplicates(subset="reference", ignore_index=True).sort_values("reference").to_excel(
            master_writer, sheet_name="citations", index=False
        )

    written_files.append(master_fp)
    for thefile in written_files:
        logger.info(f"Excel file created: {thefile}")
    return written_files


def export(
    domain: str = "enterprise-attack",
    version: str = None,
    output_dir: str = ".",
    remote: str = None,
    stix_file: str = None,
):
    """Download ATT&CK data from MITRE/CTI and convert it to Excel spreadsheets.

    Parameters
    ----------
    domain : str, optional
        The domain of ATT&CK to download, e.g "enterprise-attack", by default "enterprise-attack"
    version : str, optional
        The version of ATT&CK to download, e.g "v8.1".
        If omitted will build the current version of ATT&CK, by default None
    output_dir : str, optional
        The directory to write the excel files to.
        If omitted writes to a subfolder of the current directory depending on specified domain and version, by default "."
    remote : str, optional
        The URL of a remote ATT&CK Workbench instance to connect to for stix data.
        Mutually exclusive with stix_file.
        by default None
    stix_file : str, optional
        Path to a local STIX file containing ATT&CK data for a domain, by default None

    Raises
    ------
    ValueError
        Raised if both `remote` and `stix_file` are passed
    """
    if remote and stix_file:
        raise ValueError("remote and stix_file are mutually exclusive. Please only use one or the other")

    mem_store = get_stix_data(domain=domain, version=version, remote=remote, stix_file=stix_file)

    logger.info(f"************ Exporting {domain} to Excel ************")

    # build dataframes
    dataframes = build_dataframes(src=mem_store, domain=domain)
    write_excel(dataframes=dataframes, domain=domain, version=version, output_dir=output_dir)


def main():
    """Entrypoint for attackToExcel_cli."""
    parser = argparse.ArgumentParser(
        description="Download ATT&CK data from MITRE/CTI and convert it to excel spreadsheets"
    )
    parser.add_argument(
        "-domain",
        type=str,
        choices=["enterprise-attack", "mobile-attack", "ics-attack"],
        default="enterprise-attack",
        help="which domain of ATT&CK to convert",
    )
    parser.add_argument(
        "-version",
        type=str,
        help="which version of ATT&CK to convert. If omitted, builds the latest version",
    )
    parser.add_argument(
        "-output",
        type=str,
        default=".",
        help="output directory. If omitted writes to a subfolder of the current directory depending on "
        "the domain and version",
    )
    parser.add_argument(
        "-remote",
        type=str,
        default=None,
        help="remote url of an ATT&CK workbench server. If omitted, stix data will be acquired from the"
        " official ATT&CK Taxii server (cti-taxii.mitre.org)",
    )
    parser.add_argument(
        "-stix-file",
        type=str,
        default=None,
        help="Path to a local STIX file containing ATT&CK data for a domain, by default None",
    )
    args = parser.parse_args()

    export(
        domain=args.domain, version=args.version, output_dir=args.output, remote=args.remote, stix_file=args.stix_file
    )


if __name__ == "__main__":
    main()



"""Helper functions."""
from mitreattack.constants import MITRE_ATTACK_ID_SOURCE_NAMES


def remove_revoked_depreciated(listing):
    """Remove revoked and depreciated elements from the listing.

    :param listing: input element list
    :return: input element list - revoked elements
    """
    return list(
        filter(lambda x: x.get("x_mitre_deprecated", False) is False and x.get("revoked", False) is False, listing)
    )


def construct_relationship_mapping(mapping_obj, rel):
    """Will safely add a relationship to a listing in the mapping_obj.

    :param mapping_obj: A dictionary containing mappings of attack object types (course-of-action,
                        tool, malware, or intrusion-set) to relationships
    :param rel: The relationship to add
    :return: Updated mapping_obj
    """
    if rel["target_ref"] not in mapping_obj:
        mapping_obj[rel["target_ref"]] = []
    mapping_obj[rel["target_ref"]].append(rel)


def get_attack_id(obj):
    """Get the ATT&CK ID from an object.

    :param obj: The object to extract from
    :return: The ATT&CK ID in string form
    """
    if not obj["type"].startswith("x-mitre"):
        for entry in obj["external_references"]:
            if entry["source_name"] in MITRE_ATTACK_ID_SOURCE_NAMES:
                return entry["external_id"]
    else:
        return obj["id"]
    return "-1"


def build_data_strings(data_sources, data_components):
    """Build source->component strings for layer generation.

    :param data_sources: List of Data Sources (dicts)
    :param data_components: List of Data Components (dicts)
    :return: dict mapping of Data Component IDs to generated source->component strings
    """
    out = dict()
    for component in data_components:
        ref = component["x_mitre_data_source_ref"]
        try:
            source = [x for x in data_sources if x["id"] == ref][0]
            name = f"{source['name']}: {component['name']}"
            out[component["id"]] = name
        except IndexError:
            print(
                f'[generator] - (WARNING): no matching data source{ref} found for data component {component["id"]}. '
                f"Skipping..."
            )
            pass
    return out



"""Contains custom exceptions."""


UNSETVALUE = "(x)"


class BadInput(Exception):
    """Custom exception used for bad input."""

    pass


class BadType(Exception):
    """Custom exception used for bad types."""

    pass


class UninitializedLayer(Exception):
    """Custom exception used for uninitialized layers."""

    pass


class UnknownLayerProperty(Exception):
    """Custom exception used for unknown layer properties."""

    pass


class UnknownTechniqueProperty(Exception):
    """Custom exception used for unknown technique properties."""

    pass


class MissingParameters(Exception):
    """Custom exception used for missing parameters."""

    pass


def handler(caller, msg):
    """Print a debug/warning/error message.

    :param caller: the entity that called this function
    :param msg: the message to log
    """
    print(f"[{caller}] - {msg}")


def typeChecker(caller, testee, desired_type, field):
    """Verify that the tested object is of the correct type.

    :param caller: the entity that called this function (used for error
        messages)
    :param testee: the element to test
    :param desired_type: the type the element should be
    :param field: what the element is to be used as (used for error
        messages)
    :raises BadType: error denoting the testee element is not of the
        correct type
    """
    if not isinstance(testee, desired_type):
        handler(caller, f"{testee} [{field}] is not a {str(desired_type)}")
        raise BadType


def typeCheckerArray(caller, testee, desired_type, field):
    """Verify that the tested object is an array of the correct type.

    :param caller: the entity that called this function (used for error
        messages)
    :param testee: the element to test
    :param desired_type: the type the element should be
    :param field: what the element is to be used as (used for error
        messages)
    :raises BadType: error denoting the testee element is not of the
        correct type
    """
    if not isinstance(testee, list):
        handler(caller, f"{testee} [{field}] is not a Array")
        raise BadType
    if not isinstance(testee[0], desired_type):
        handler(caller, f"{testee} [{field}] is not a {'Array of ' + desired_type}")
        raise BadType


def categoryChecker(caller, testee, valid, field):
    """Verify that the tested object is one of a set of valid values.

    :param caller: the entity that called this function (used for error
        messages)
    :param testee: the element to test
    :param valid: a list of valid values for the testee
    :param field: what the element is to be used as (used for error
        messages)
    :raises BadInput: error denoting the testee element is not one of
        the valid options
    """
    if testee not in valid:
        handler(caller, f"{testee} not a valid value for {field}")
        raise BadInput


def loadChecker(caller, testee, required, field):
    """Verify that the tested object contains all required fields.

    :param caller: the entity that called this function (used for error
        messages)
    :param testee: the element to test
    :param required: a list of required values for the testee
    :param field: what the element is to be used as (used for error
        messages)
    :raises BadInput: error denoting the testee element is not one of
        the valid options
    """
    for entry in required:
        if entry not in testee:
            handler(caller, f"{entry} is not present in {field} [{testee}]")
            raise MissingParameters



"""Contains LayerOps class."""

# Example Use:
# from navlayers.manipulators.layerops import LayerOps
# from navlayers.core.layer import Layer

# demo = Layer()
# demo.load_file("C:\Users\attack\Downloads\layer.json")
# demo2 = Layer()
# demo2.load_input(Existing_Layer_String_Previously_Loaded)

# lo = LayerOps(score=lambda x: x[0] * x[1], name=lambda x: x[1],
#               desc=lambda x: "This is an list example")
# out_layer = lo.process([demo, demo2])
# out_layer.export_file("C:\demo_layer1.json")

# lo2 = LayerOps(score=lambda x: x['a'], color=lambda x: x['b'],
#                desc=lambda x: "This is a dict example")
# out_layer2 = lo2.process({'a': demo, 'b': demo2})
# dict_layer = out_layer2.get_dict()
# print(dict_layer)
# out_layer2.export_file("C:\demo_layer2.json")

import copy
from mitreattack.navlayers.core import Layer


class InvalidFormat(Exception):
    """Custom exception used for invalid formats."""

    pass


class BadLambda(Exception):
    """Custom exception used for bad lambdas."""

    pass


class MismatchedDomain(Exception):
    """Custom exception used for mismatched domains."""

    pass


class LayerOps:
    """A LayerOps object."""

    def __init__(
        self,
        score=None,
        comment=None,
        enabled=None,
        colors=None,
        metadata=None,
        name=None,
        desc=None,
        default_values=None,
    ):
        """Initialize - configures the object to handle processing based on user provided Lambdas.

        :param score: lambda to calculate score
        :param comment: lambda to generate comments
        :param enabled: lambda to determine enabled status
        :param metadata: lambda to generate metadata
        :param name: new name to apply to the resulting layer
        :param desc: new description to apply to the resulting layer
        :param default_values: dictionary containing desired default
            values for missing data element values
        """
        self._score = score
        self._comment = comment
        self._enabled = enabled
        self._colors = colors
        self._metadata = metadata
        self._name = name
        self._desc = desc
        self._default_values = {"comment": "", "enabled": True, "color": "#ffffff", "score": 1, "metadata": []}
        if default_values is not None:
            for entry in default_values:
                self._default_values[entry] = default_values[entry]

    def process(self, data, default_values=None):
        """Take a list or dict of Layer objects, and processes them.

        :param data: A dict or list of Layer objects.
        :param default_values: dictionary containing desired default values for
            missing data element values
        :raises InvalidFormat: An error indicating that the layer data
            wasn't provided in a list or dict
        """
        if isinstance(data, dict):
            temp = {}
            for entry in data.keys():
                temp[entry] = data[entry].layer.techniques
            self.mode = "dict"
            out = self._merge_to_template(data, key=list(data.keys())[0])
        elif isinstance(data, list):
            temp = []
            for entry in data:
                temp.append(entry.layer.techniques)
            self.mode = "list"
            out = self._merge_to_template(data)
        else:
            raise InvalidFormat
        da = temp
        corpus = self._build_template(temp)

        defaults = self._default_values
        if default_values is not None:
            for entry in default_values:
                defaults[entry] = default_values[entry]

        return self._compute(data, da, corpus, out, defaults)

    def _compute(self, data, da, corpus, out, defaults):
        """Apply the configured lambda to the dataset.

        :param data: the dataset being processed
        :param da: extracted techniques from the dataset, sorted by
            dataset format
        :param corpus: master list of combined techniques and
            technique data
        :param out: baseline template for the new layer
        :param defaults: default values each technique should use if a
            field is missing
        :returns: a Layer object representing the resultant layer
        """
        composite = copy.deepcopy(corpus)
        if self._score is not None:
            for entry in composite:
                entry["score"] = self._applyOperation(da, entry, "score", self._score, defaults)

        if self._comment is not None:
            for entry in composite:
                entry["comment"] = self._applyOperation(da, entry, "comment", self._comment, defaults)

        if self._enabled is not None:
            for entry in composite:
                entry["enabled"] = self._applyOperation(da, entry, "enabled", self._enabled, defaults)

        if self._colors is not None:
            for entry in composite:
                entry["color"] = self._applyOperation(da, entry, "color", self._colors, defaults)

        if self._metadata is not None:
            for entry in composite:
                entry["metadata"] = self._applyOperation(da, entry, "metadata", self._metadata, defaults)

        processed = copy.deepcopy(out)
        processed["techniques"] = composite
        if self._name is not None:
            processed["name"] = self._applyOperation(data, None, "name", self._name, defaults, glob="name")

        if self._desc is not None:
            processed["description"] = self._applyOperation(
                data, None, "description", self._desc, defaults, glob="description"
            )

        return Layer(processed)

    def _merge_to_template(self, data, key=0):
        """Merge initial layer files in either dict or list form into a single template. Defaults to the first entry in the case of difference in metadata.

        :param key: the key referencing the first entry to default to
        :raises MismatchedDomain: An error indicating that the layers
            came from different domains
        """
        out = {}
        dict_map = []
        collide = []
        if self.mode == "dict":
            for x in data.keys():
                dict_map.append(x)
                collide.append(data[x])
        else:
            for x in data:
                collide.append(x)
        key_space = data[key].layer._enumerate()
        _raw = data[key].layer.get_dict()
        for entry in key_space:
            if entry != "techniques":
                standard = _raw[entry]
                if any(y != standard for y in [getattr(x.layer, entry) for x in collide]):
                    if entry == "domain":
                        print("FATAL ERROR! Layer mis-match on domain. " "Exiting.")
                        raise MismatchedDomain
                    print(f"Warning! Layer mis-match detected for {entry}. Defaulting to {key}'s value")
                out[entry] = standard
        return out

    def _build_template(self, data):
        """Build a base template by combining available technique listings from each layer.

        :param data: the raw ingested technique data (list or dict)
        """
        if self.mode == "list":
            return self._template(data)
        elif self.mode == "dict":
            temp = {}
            t2 = []
            for key in data:
                temp[key] = self._template([data[key]])
            for key in temp:
                for elm in temp[key]:
                    if not any(elm["techniqueID"] == x["techniqueID"] and elm["tactic"] == x["tactic"] for x in t2):
                        t2.append(elm)
                    else:
                        [
                            x.update(elm)
                            if x["techniqueID"] == elm["techniqueID"] and elm["tactic"] == x["tactic"]
                            else x
                            for x in t2
                        ]
            return t2

    @staticmethod
    def _template(data):
        """Create a template technique entry for a given listing of techniques.

        :param data: a single layer's technique data
        :returns: a list of technique templates
        """
        temp = []
        for entry in data:
            temp.append(
                [
                    {"techniqueID": x.techniqueID, "tactic": x.tactic} if x.tactic else {"techniqueID": x.techniqueID}
                    for x in entry
                ]
            )
        complete = []
        for entry in temp:
            for val in entry:
                if val in complete:
                    continue
                complete.append(val)
        return complete

    def _grabList(self, search, collection):
        """Generate a list containing all values for a given key across the collection.

        :param search: the key to search for
        :param collection: the data collection to search
        :returns: a list of values for that key across the collection
        """
        temp = []
        for x in collection:
            temp.append(self._grabElement(search, x))
        return temp

    def _grabDict(self, search, collection):
        """Generate a dictionary containing all values for a given key across the collection.

        :param search: the key to search for
        :param collection: the data collection to search
        :returns: a dict of values for that key across the collection
        """
        temp = {}
        for key in collection:
            temp[key] = self._grabElement(search, collection[key])
        return temp

    def _grabElement(self, search, listing):
        """Return a matching element in the listing for the search key.

        :param search: the key to search for
        :param listing: the data element to search
        :returns: the found value, or an empty dict
        """
        val = [x for x in listing if self._inDict(search, x)]
        if len(val) > 0:
            return val[0]
        return {}

    @staticmethod
    def _inDict(search, complete):
        """Return bool of whether or not the key searched for can be found across the dataset corpus.

        :param search: the key to search for
        :param complete: the data set to search for
        :returns: true/false
        """
        comp_list = complete.get_dict().items()
        search_terms = search.items()
        return all(elm in comp_list for elm in search_terms)

    def _applyOperation(self, corpus, element, name, lda, defaults, glob=None):
        """Apply a lambda expression to the dataset.

        :param corpus: the dataset
        :param element: the template file to fill out
        :param name: the name of the field being processed
        :param lda: the lambda expression to apply
        :param defaults: any default values should the field not be found
            in a dataset entry
        :raises BadLambda: error denoting that an error has occurred when
            running the lambda function
        :returns: lambda output
        """
        if self.mode == "list":
            if glob:
                listing = [getattr(x.layer, glob) for x in corpus]
                listing = [{name: x} for x in listing]
            else:
                te = dict(techniqueID=element["techniqueID"])
                if "tactic" in element:
                    te["tactic"] = element["tactic"]
                listing = self._grabList(te, corpus)
                listing = [x.get_dict() if not isinstance(x, dict) else dict() for x in listing]
            values = []
            for x in listing:
                if name in x:
                    values.append(x[name])
                else:
                    if defaults is not None:
                        if name in defaults:
                            values.append(defaults[name])
                            continue
                    values.append(self._default_values[name])
        else:
            values = {}
            if glob:
                listing = {}
                for k in corpus.keys():
                    listing[k] = {glob: getattr(corpus[k].layer, glob)}
            else:
                te = dict(techniqueID=element["techniqueID"])
                if "tactic" in element:
                    te["tactic"] = element["tactic"]
                temp = self._grabDict(te, corpus)
                listing = {}
                for k in temp.keys():
                    if temp[k] != {}:
                        listing[k] = temp[k].get_dict()
                    else:
                        listing[k] = {}
            for elm in listing:
                if name in listing[elm]:
                    values[elm] = listing[elm][name]
                else:
                    if defaults is not None:
                        if name in defaults:
                            values[elm] = defaults[name]
                            continue
                    values[elm] = self._default_values[name]
        try:
            return lda(values)
        except IndexError and KeyError:
            print(
                'Unable to continue, lambda targeting "{}" could not operate'
                " correctly on {}. Maybe the field is missing?".format(name, element)
            )
            print(f"[RAW] Extracted matching elements: {listing}")
            raise BadLambda



"""Contains Versions class."""

from mitreattack.navlayers.core.exceptions import typeChecker, categoryChecker, UNSETVALUE, BadInput

defaults = dict(layer="4.3", navigator="4.5.5")


class Versions:
    """A Versions object."""

    def __init__(self, layer=defaults["layer"], attack=UNSETVALUE, navigator=defaults["navigator"]):
        """Initialize - Creates a v4 Versions object.

        :param layer: The layer version
        :param attack: The attack version
        :param navigator: The navigator version
        """
        self.layer = layer
        self.__attack = attack
        self.navigator = navigator

    @property
    def attack(self):
        """Getter for attack."""
        if self.__attack != UNSETVALUE:
            return self.__attack
        else:
            return "4.x"

    @attack.setter
    def attack(self, attack):
        """Setter for attack."""
        typeChecker(type(self).__name__, attack, str, "attack")
        self.__attack = attack

    @property
    def navigator(self):
        """Getter for navigator."""
        return self.__navigator

    @navigator.setter
    def navigator(self, navigator):
        """Setter for navigator."""
        typeChecker(type(self).__name__, navigator, str, "navigator")
        if not navigator.startswith("4."):
            print(
                f"[WARNING] - unrecognized navigator version {navigator}. Defaulting to the 4.X schema, "
                f"this may result in unexpected behavior."
            )
            navigator = defaults["navigator"]
        self.__navigator = navigator

    @property
    def layer(self):
        """Getter for layer."""
        return self.__layer

    @layer.setter
    def layer(self, layer):
        """Setter for layer."""
        typeChecker(type(self).__name__, layer, str, "layer")
        try:
            categoryChecker(type(self).__name__, layer, ["3.0", "4.0", "4.1", "4.2", "4.3"], "layer version")
        except BadInput:
            print(
                f"[WARNING] - unrecognized layer version {layer}. Defaulting to the 4.3 schema, this may result in "
                f"unexpected behavior."
            )
        if layer in ["3.0", "4.0", "4.1", "4.2"]:
            print(f"[NOTICE] - Forcibly upgrading version from {layer} to 4.3.")
            layer = "4.3"
        self.__layer = layer

    def get_dict(self):
        """Convert the currently loaded data into a dict.

        :returns: A dict representation of the local Versions object
        """
        temp = dict()
        listing = vars(self)
        for entry in listing:
            if listing[entry] != UNSETVALUE:
                subname = entry.split("__")[-1]
                temp[subname] = listing[entry]
        return temp



"""Contains SVGConfig and ToSvg classes."""

import json
from copy import deepcopy

from mitreattack.navlayers.core import Layer
from mitreattack.navlayers.exporters.svg_templates import SvgTemplates


class NoLayer(Exception):
    """Custom exception used when no layer is found."""

    pass


class SVGConfig:
    """A SVGConfig object."""

    d_width = 8.5
    d_height = 11
    d_headerHeight = 1
    d_unit = "in"
    d_showSubtechniques = "expanded"
    d_font = "sans-serif"
    d_tableBorderColor = "#6B7279"
    d_showHeader = True
    d_legendDocked = True
    d_legendX = 0
    d_legendY = 0
    d_legendWidth = 2
    d_legendHeight = 1
    d_showLegend = True
    d_showFilters = True
    d_showAbout = True
    d_showDomain = True
    d_border = 0.104

    def __init__(
        self,
        width=d_width,
        height=d_height,
        headerHeight=d_headerHeight,
        unit=d_unit,
        showSubtechniques=d_showSubtechniques,
        font=d_font,
        tableBorderColor=d_tableBorderColor,
        showHeader=d_showHeader,
        legendDocked=d_legendDocked,
        legendX=d_legendX,
        legendY=d_legendY,
        legendWidth=d_legendWidth,
        legendHeight=d_legendHeight,
        showLegend=d_showLegend,
        showFilters=d_showFilters,
        showAbout=d_showAbout,
        showDomain=d_showDomain,
        border=d_border,
    ):
        """Define parameters to configure SVG export.

        :param width: Desired SVG width
        :param height: Desired SVG height
        :param headerHeight: Desired Header Block height
        :param unit: SVG measurement units (qualifies width, height, etc.) - "in", "cm", "px", "em", or "pt"
        :param showSubtechniques: Display form for subtechniques - "all", "expanded" (decided by layer), or "none"
        :param font: What font style to use - "serif", "sans-serif", or "monospace"
        :param tableBorderColor: Hex color to use for the technique borders
        :param showHeader: Whether or not to show Header Blocks
        :param legendDocked: Whether or not the legend should be docked
        :param legendX: Where to place the legend on the x axis if not docked
        :param legendY: Where to place the legend on the y axis if not docked
        :param legendWidth: Width of the legend if not docked
        :param legendHeight: Height of the legend if not docked
        :param showLegend: Whether or not to show the legend
        :param showFilters: Whether or not to show the Filter Header Block
        :param showAbout: Whether or not to show the About Header Block
        :param showDomain: Whether or not to show the Domain Version Header Block
        :param border: What default border width to use
        """
        # force defaults in case bad values are provided so we don't crash later
        self.width = self.d_width
        self.height = self.d_height
        self.headerHeight = self.d_headerHeight
        self.unit = self.d_unit
        self.showSubtechniques = self.d_showSubtechniques
        self.font = self.d_font
        self.tableBorderColor = self.d_tableBorderColor
        self.showHeader = self.d_showHeader
        self.legendDocked = self.d_legendDocked
        self.legendX = self.d_legendX
        self.legendY = self.d_legendY
        self.legendWidth = self.d_legendWidth
        self.legendHeight = self.d_legendHeight
        self.showDomain = self.d_showDomain
        self.showLegend = self.d_showLegend
        self.showFilters = self.d_showFilters
        self.showAbout = self.d_showAbout
        self.border = self.d_border

        self.width = width
        self.height = height
        self.headerHeight = headerHeight
        self.unit = unit
        self.showSubtechniques = showSubtechniques
        self.font = font
        self.tableBorderColor = tableBorderColor
        self.showHeader = showHeader
        self.legendDocked = legendDocked
        self.legendX = legendX
        self.legendY = legendY
        self.legendWidth = legendWidth
        self.legendHeight = legendHeight
        self.showDomain = showDomain
        self.showLegend = showLegend
        self.showFilters = showFilters
        self.showAbout = showAbout
        self.border = border

    def load_from_file(self, filename=""):
        """Load config from a json file.

        :param filename: The file to read from
        """
        with open(filename, "r") as fio:
            raw = fio.read()
        self._data = json.loads(raw)
        for entry in self._data:
            patched = entry
            if not patched.startswith("_SVGConfig__"):
                patched = "_SVGConfig__" + patched
            if patched in vars(self).keys():
                setattr(self, entry, self._data[entry])
            else:
                print(f"WARNING - Unidentified Config Field in {filename}: {entry}")

        self.__str__()

    def save_to_file(self, filename=""):
        """Store config to json file.

        :param filename: The file to write to
        """
        out = dict(
            width=self.width,
            height=self.height,
            headerHeight=self.headerHeight,
            unit=self.unit,
            showSubtechniques=self.showSubtechniques,
            font=self.font,
            tableBorderColor=self.tableBorderColor,
            showHeader=self.showHeader,
            legendDocked=self.legendDocked,
            legendX=self.legendX,
            legendY=self.legendY,
            legendWidth=self.legendWidth,
            legendHeight=self.legendHeight,
            showLegend=self.showLegend,
            showFilters=self.showFilters,
            showAbout=self.showAbout,
            border=self.border,
        )
        with open(filename, "w", encoding="utf-16") as file:
            json.dump(out, file, ensure_ascii=False)

    def __str__(self):
        """Display current configuration."""
        print("SVGConfig current settings: ")
        print(f"width - {self.width}")
        print(f"height - {self.height}")
        print(f"headerHeight - {self.headerHeight}")
        print(f"unit - {self.unit}")
        print(f"showSubtechniques - {self.showSubtechniques}")
        print(f"font - {self.font}")
        print(f"tableBorderColor - {self.tableBorderColor}")
        print(f"showHeader - {self.showHeader}")
        print(f"legendDocked - {self.legendDocked}")
        print(f"legendX - {self.legendX}")
        print(f"legendY - {self.legendY}")
        print(f"legendWidth - {self.legendWidth}")
        print(f"legendHeight - {self.legendHeight}")
        print(f"showLegend - {self.showLegend}")
        print(f"showFilters - {self.showFilters}")
        print(f"showAbout - {self.showAbout}")
        print(f"border - {self.border}")

    @property
    def width(self):
        """Width getter."""
        if self.__width is not None:
            return self.__width

    @width.setter
    def width(self, width):
        """Width setter."""
        if isinstance(width, int) or isinstance(width, float):
            self.__width = width
        else:
            print(f"[Warning] - Unable to set width to {width}: not a float or int")

    @property
    def height(self):
        """Height getter."""
        if self.__height is not None:
            return self.__height

    @height.setter
    def height(self, height):
        """Height setter."""
        if isinstance(height, int) or isinstance(height, float):
            self.__height = height
        else:
            print(f"[Warning] - Unable to set height to {height}: not a float or int")

    @property
    def headerHeight(self):
        """Header Height getter."""
        if self.__headerHeight is not None:
            return self.__headerHeight

    @headerHeight.setter
    def headerHeight(self, headerHeight):
        """Header Height setter."""
        if isinstance(headerHeight, int) or isinstance(headerHeight, float):
            self.__headerHeight = headerHeight
        else:
            print(f"[Warning] - Unable to set headerHeight to {headerHeight}: not a float or int")

    @property
    def unit(self):
        """Getter for Unit."""
        if self.__unit is not None:
            return self.__unit

    @unit.setter
    def unit(self, unit):
        """Setter for Unit."""
        if unit in ["in", "cm", "px", "em", "pt"]:
            self.__unit = unit
        else:
            print(f'[Warning] - Unable to set unit to {unit}: not one of ["in", "cm", "px", "em", "pt"]')

    @property
    def showSubtechniques(self):
        """Show Subtechniques getter."""
        if self.__showSubtechniques is not None:
            return self.__showSubtechniques

    @showSubtechniques.setter
    def showSubtechniques(self, showSubtechniques):
        """Show Subtechniques setter."""
        if showSubtechniques in ["expanded", "all", "none"]:
            self.__showSubtechniques = showSubtechniques
        else:
            print(
                '[Warning] - Unable to set showSubtechniques to {}: not one of ["expanded", "all", "none"]'.format(
                    showSubtechniques
                )
            )

    @property
    def font(self):
        """Font getter."""
        if self.__font is not None:
            return self.__font

    @font.setter
    def font(self, font):
        """Font setter."""
        if font in ["serif", "sans-serif", "monospace"]:
            self.__font = font
        else:
            print(f'[Warning] - Unable to set font to {font}: not one of ["serif", "sans-serif", "monospace"]')

    @property
    def tableBorderColor(self):
        """Table Border Color getter."""
        if self.__tableBorderColor is not None:
            return self.__tableBorderColor

    @tableBorderColor.setter
    def tableBorderColor(self, tableBorderColor):
        """Table Border Color setter."""
        if isinstance(tableBorderColor, str) and tableBorderColor.startswith("#") and len(tableBorderColor) == 7:
            self.__tableBorderColor = tableBorderColor
        else:
            reason = ""
            if not isinstance(tableBorderColor, str):
                reason = "not a string"
            elif not tableBorderColor.startswith("#"):
                reason = "not a valid code (does not start with #)"
            elif len(tableBorderColor) != 7:
                reason = "not a valid code (#ZZZZZZ)"
            print(f"[Warning] - Unable to set tableBorderColor to {tableBorderColor}: " + reason)

    @property
    def showHeader(self):
        """Show Header getter."""
        if self.__showHeader is not None:
            return self.__showHeader

    @showHeader.setter
    def showHeader(self, showHeader):
        """Show Header setter."""
        if isinstance(showHeader, bool):
            self.__showHeader = showHeader
        else:
            print(f"[Warning] - Unable to set showHeader to {showHeader}: not a bool")

    @property
    def legendDocked(self):
        """Legend Docked getter."""
        if self.__legendDocked is not None:
            return self.__legendDocked

    @legendDocked.setter
    def legendDocked(self, legendDocked):
        """Legend Docked setter."""
        if isinstance(legendDocked, bool):
            self.__legendDocked = legendDocked
        else:
            print(f"[Warning] - Unable to set legendDocked to {legendDocked}: not a bool")

    @property
    def legendX(self):
        """Legend X getter."""
        if self.__legendX is not None:
            return self.__legendX

    @legendX.setter
    def legendX(self, legendX):
        """Legend X setter."""
        if isinstance(legendX, int) or isinstance(legendX, float):
            self.__legendX = legendX
        else:
            print(f"[Warning] - Unable to set legendX to {legendX}: not a float or int")

    @property
    def legendY(self):
        """Legend Y getter."""
        if self.__legendY is not None:
            return self.__legendY

    @legendY.setter
    def legendY(self, legendY):
        """Legend Y setter."""
        if isinstance(legendY, int) or isinstance(legendY, float):
            self.__legendY = legendY
        else:
            print(f"[Warning] - Unable to set legendY to {legendY}: not a float or int")

    @property
    def legendWidth(self):
        """Legend Width getter."""
        if self.__legendWidth is not None:
            return self.__legendWidth

    @legendWidth.setter
    def legendWidth(self, legendWidth):
        """Legend Width setter."""
        if isinstance(legendWidth, int) or isinstance(legendWidth, float):
            self.__legendWidth = legendWidth
        else:
            print(f"[Warning] - Unable to set legendWidth to {legendWidth}: not a float or int")

    @property
    def legendHeight(self):
        """Legend Height getter."""
        if self.__legendHeight is not None:
            return self.__legendHeight

    @legendHeight.setter
    def legendHeight(self, legendHeight):
        """Legend Height setter."""
        if isinstance(legendHeight, int) or isinstance(legendHeight, float):
            self.__legendHeight = legendHeight
        else:
            print(f"[Warning] - Unable to set legendHeight to {legendHeight}: not a float or int")

    @property
    def showLegend(self):
        """Show Legend getter."""
        if self.__showLegend is not None:
            return self.__showLegend

    @showLegend.setter
    def showLegend(self, showLegend):
        """Show Legend setter."""
        if isinstance(showLegend, bool):
            self.__showLegend = showLegend
        else:
            print(f"[Warning] - Unable to set showLegend to {showLegend}: not a bool")

    @property
    def showFilters(self):
        """Show Filters getter."""
        if self.__showFilters is not None:
            return self.__showFilters

    @showFilters.setter
    def showFilters(self, showFilters):
        """Show Filters setter."""
        if isinstance(showFilters, bool):
            self.__showFilters = showFilters
        else:
            print(f"[Warning] - Unable to set showFilters to {showFilters}: not a bool")

    @property
    def showAbout(self):
        """Show About getter."""
        if self.__showAbout is not None:
            return self.__showAbout

    @showAbout.setter
    def showAbout(self, showAbout):
        """Show About setter."""
        if isinstance(showAbout, bool):
            self.__showAbout = showAbout
        else:
            print(f"[Warning] - Unable to set showAbout to {showAbout}: not a bool")

    @property
    def showDomain(self):
        """Show Domain getter."""
        if self.__showDomain is not None:
            return self.__showDomain

    @showDomain.setter
    def showDomain(self, showDomain):
        """Show Domain setter."""
        if isinstance(showDomain, bool):
            self.__showDomain = showDomain
        else:
            print(f"[Warning] - Unable to set showAbout to {showDomain}: not a bool")

    @property
    def border(self):
        """Border getter."""
        if self.__border is not None:
            return self.__border

    @border.setter
    def border(self, border):
        """Border setter."""
        if isinstance(border, float):
            self.__border = border
        else:
            print(f"[Warning] - Unable to set border to {border}: not a float")


class ToSvg:
    """A ToSvg object."""

    def __init__(self, domain="enterprise", source="taxii", resource=None, config=None):
        """Set up exporting system, builds underlying matrix.

        :param domain: Which domain to utilize for the underlying matrix layout
        :param source: Use the taxii server, local data, or a remote ATT&CK Workbench instance
        :param resource: string path to local cache of stix data (local) or url of workbench to reach out
                            to (remote)
        :param config: Optional pre-existing SVGConfig object
        """
        self.raw_handle = SvgTemplates(domain=domain, source=source, resource=resource)
        if config is not None and isinstance(config, SVGConfig):
            self.config = config
        else:
            self.config = SVGConfig()

    def to_svg(self, layerInit, filepath="example.svg"):
        """Generate a svg file based on the input layer.

        :param layerInit: Input attack layer object to transform
        :param filepath: Desired output svg location
        :return: (meta) svg file at the targeted output location
        """
        if layerInit is not None:
            if not isinstance(layerInit, Layer):
                raise TypeError

        if layerInit is None:
            raise NoLayer

        layer = deepcopy(layerInit)

        included_subs = []
        if layer.layer.techniques:
            for entry in layer.layer.techniques:
                if self.config.showSubtechniques == "expanded":
                    if entry.showSubtechniques:
                        if entry.tactic:
                            included_subs.append((entry.techniqueID, entry.tactic))
                        else:
                            included_subs.append((entry.techniqueID, False))
                elif self.config.showSubtechniques == "all":
                    if entry.tactic:
                        included_subs.append((entry.techniqueID, entry.tactic))
                    else:
                        included_subs.append((entry.techniqueID, False))
                else:  # none displayed
                    pass

        excluded = []
        if layer.layer.hideDisabled:
            for entry in layer.layer.techniques:
                if entry.enabled is False:
                    if entry.tactic:
                        excluded.append((entry.techniqueID, entry.tactic))
                    else:
                        excluded.append((entry.techniqueID, False))
        scores = []
        colors = []
        if layer.layer.techniques:
            for entry in layer.layer.techniques:
                tscore = entry.score
                if tscore is not None:
                    if entry.tactic:
                        scores.append((entry.techniqueID, entry.tactic, tscore))
                    else:
                        scores.append((entry.techniqueID, False, tscore))
                elif entry.color:
                    if entry.tactic:
                        colors.append((entry.techniqueID, entry.tactic, entry.color))
                    else:
                        colors.append((entry.techniqueID, False, entry.color))
        sName = True
        sID = False
        sort = 0
        if layer.layer.layout:
            sName = layer.layer.layout.showName
            sID = layer.layer.layout.showID
        if layer.layer.sorting:
            sort = layer.layer.sorting
        d = self.raw_handle.export(
            showName=sName,
            showID=sID,
            sort=sort,
            scores=scores,
            subtechs=included_subs,
            colors=colors,
            exclude=excluded,
            lhandle=layer.layer,
            config=self.config,
        )
        d.saveSvg(filepath)



"""Entrypoint for the layer exporter cli."""

import argparse

from mitreattack.navlayers.exporters.to_svg import ToSvg, SVGConfig
from mitreattack.navlayers.exporters.to_excel import ToExcel
from mitreattack.navlayers.core import Layer


def main(argv=None):
    """Entrypoint for layerExporter_cli."""
    parser = argparse.ArgumentParser(description="Export an ATT&CK Navigator layer as a svg image or excel file")
    parser.add_argument(
        "-m", "--mode", choices=["svg", "excel"], required=True, help="The form to export the layers in"
    )
    parser.add_argument("input", nargs="+", help="Path(s) to the file to export")
    parser.add_argument(
        "-s",
        "--source",
        choices=["taxii", "local", "remote"],
        default="taxii",
        help="What source to utilize when building the matrix",
    )
    parser.add_argument(
        "--resource",
        help="Path to the local resource if --source=local, or url of an ATT&CK Workbench"
        " instance if --source=remote",
        default=None,
    )
    parser.add_argument("-o", "--output", nargs="+", help="Path(s) to the exported svg/xlsx file", required=True)
    parser.add_argument(
        "-l",
        "--load_settings",
        help="[SVG Only] Path to a SVG configuration json to use when " "rendering",
        default=None,
    )
    parser.add_argument(
        "-d",
        "--size",
        nargs=2,
        help="[SVG Only] X and Y size values (in inches) for SVG export (use " "-l for other settings)",
        default=[8.5, 11],
        metavar=("WIDTH", "HEIGHT"),
    )
    args = parser.parse_args(args=argv)

    if len(args.output) != len(args.input):
        print("Mismatched number of output file paths to input file paths. Exiting...")
        return

    for i in range(0, len(args.input)):
        entry = args.input[i]
        print(f"{i + 1}/{len(args.input)} - Beginning processing {entry}")
        lay = Layer()
        try:
            lay.from_file(entry)
        except Exception as e:
            print(f"Unable to load {entry} due to exception: {e}. Skipping...")
            continue
        if args.mode == "excel":
            if not args.output[i].endswith(".xlsx"):
                print(f"[ERROR] Unable to export {entry} as type: excel to {args.output[i]}")
                continue
            exy = ToExcel(domain=lay.layer.domain, source=args.source, resource=args.resource)
            exy.to_xlsx(lay, filepath=args.output[i])
        else:
            if not args.output[i].endswith(".svg"):
                print(f"[ERROR] Unable to export {entry} as type: svg to {args.output[i]}")
                continue
            conf = SVGConfig()
            if args.load_settings:
                conf.load_from_file(args.load_settings)
            if len(args.size) == 2:
                conf.width = float(args.size[0])
                conf.height = float(args.size[1])
            svy = ToSvg(domain=lay.layer.domain, source=args.source, resource=args.resource, config=conf)
            svy.to_svg(lay, filepath=args.output[i])
        print(f"{i + 1}/{len(args.input)} - Finished exporting {entry} to {args.output[i]}")


if __name__ == "__main__":
    main()



"""Entrypoint for the layer generator cli."""

import argparse
import os

from mitreattack.navlayers.generators.overview_generator import OverviewLayerGenerator
from mitreattack.navlayers.generators.usage_generator import UsageLayerGenerator
from mitreattack.navlayers.generators.sum_generator import BatchGenerator


def main(argv=None):
    """Entrypoint for layerGenerator_cli."""
    parser = argparse.ArgumentParser(description="Generate an ATT&CK Navigator layer")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--overview-type",
        choices=["group", "software", "mitigation", "datasource"],
        help="Output a layer file where the target type is summarized across the entire dataset.",
    )
    group.add_argument(
        "--mapped-to",
        help="Output layer file with techniques mapped to the given group, software, "
        "mitigation, or data component. Argument can be name, associated "
        "group/software, or ATT&CK ID.",
    )
    group.add_argument(
        "--batch-type",
        choices=["group", "software", "mitigation", "datasource"],
        help="Output a collection of layer files to the specified folder, each one representing a "
        "different instance of the target type.",
    )
    parser.add_argument("-o", "--output", help="Path to the output layer file/directory", default="generated_layer")
    parser.add_argument(
        "--domain", help="Which domain to build off of", choices=["enterprise", "mobile", "ics"], default="enterprise"
    )
    parser.add_argument(
        "--source",
        choices=["taxii", "local", "remote"],
        default="taxii",
        help="What source to utilize when building the layer files",
    )
    parser.add_argument(
        "--resource",
        help="Path to the local resource if --source=local, or url of an ATT&CK Workbench"
        " instance if --source=remote",
        default=None,
    )
    args = parser.parse_args(args=argv)

    if args.overview_type:
        og = OverviewLayerGenerator(source=args.source, domain=args.domain, resource=args.resource)
        generated = og.generate_layer(obj_type=args.overview_type)
        print("Generating Layer File")
        out_path = args.output
        if out_path == "generated_layer":
            out_path += ".json"
        generated.to_file(out_path)
        print(f"Layer file generated as {out_path}.")
    elif args.mapped_to:
        ug = UsageLayerGenerator(source=args.source, domain=args.domain, resource=args.resource)
        generated = ug.generate_layer(match=args.mapped_to)
        print("Generating Layer File")
        out_path = args.output
        if out_path == "generated_layer":
            out_path += ".json"
        generated.to_file(out_path)
        print(f"Layer file generated as {out_path}.")
    elif args.batch_type:
        bg = BatchGenerator(source=args.source, domain=args.domain, resource=args.resource)
        generated = bg.generate_layers(layers_type=args.batch_type)
        out_path = args.output
        if out_path == "generated_layer":
            out_path += "s"
        if not os.path.exists(out_path):
            os.makedirs(out_path)
        for sid in generated:
            generated[sid].to_file(f"{out_path}/{sid}.json")
        print(f"Files saved to {out_path}/")


if __name__ == "__main__":
    main()
